#lang racket
(require rackunit)
#|
(define (member? x l)
  (match l
    [(list)
      ; input = (list)
      #f
    ]
    [(list h l ...)
      ; input: l=(list "a" "b" "c" "d") x="b"
      ; h = "a"
      ; l = (list "b" "c" "d")
      (define result (member? x l))
      ; result = #t
      (or
        (equal? x h) ; here x is the first element of the list
        result       ; otherwise, x is in the rest of the list
      )
      ; expected output: #t
      ]
  )
)

;; (display "d is in the list: ")(displayln (member? "a" (list "a" "b" "c" "d")))
;; (display "d is in the list (middle): ")(displayln (member? "b" (list "a" "b" "c" "d")))
;; (display "f is *not* in the list: ")(displayln (member? "f" (list "a" "b" "c" "d")))


(define (match-prefix? prefix l)
  (match l
    [(list) #f]
    [(list h l ...)
      ; input: l=(list "foo" "Racket") prefix="R"
      ; h = "foo"
      ; l = (list "Racket")
      (define result (match-prefix? prefix l))
      ; result = #t
      (or
        (string-prefix? h prefix)
        result
      )
    ]
  )
)

(match-prefix? "R" (list "foo" "Racket"))
(match-prefix? "R" (list "foo" "bar"))
|#


(define (search f l)
  (match l
    [(list) #f]
    [(list h l ...)
      (define result (search f l))
      (or
        (f h)
        result
      )
    ]
  )
)
(define (match-prefix? prefix l)
  (search (lambda (x) (string-prefix? x prefix)) l))
(define (member? x l)
  (search (lambda (y) (equal? x y)) l))

;; (match-prefix? "R" (list "foo" "Racket"))
;; (match-prefix? "R" (list "foo" "bar"))


;; (display "d is in the list: ")(displayln (member? "a" (list "a" "b" "c" "d")))
;; (display "d is in the list (middle): ")(displayln (member? "b" (list "a" "b" "c" "d")))
;; (display "f is *not* in the list: ")(displayln (member? "f" (list "a" "b" "c" "d")))


(define (remove x l)
  (match l
    [(list)
     ; input = '()
     (list)
    ]
    [(list h l ...)
      ; input = (list 1 2 3)
      ; x = 5
      ; h = 1
      ; l = (list 2 3)
      (define result (remove x l))
      ; result = (list 2 3)
      ; expected output = (list 1 2 3)
      (cond [(equal? x h) result] ; do not copy h
            [else
              ; otherwise, copy the element back
              (cons h result)]
      )
    ]
  )
)
#|
(remove 1 (list 1 2 3))    ; (list 2 3)
(remove 10 (list 1 2 3))   ; (list 1 2 3)
(remove 1 (list))          ; (list)|#

(define (filter keep? l)
  (match l
    [(list) (list)]
    [(list h l ...) #:when (keep? h)
      (cons h (filter keep? l))]
    [(list _ l ...) (filter keep? l)]))

;; (filter even?
;;     (list 1 2 3 4 5 6 7)
;; )

(define (append l1 l2)
  (match l1
    [(list) l2]
    [(list h l1 ...)
      ; input: l1= (list 1 2) l2=(list 3 4)
      ; h = 1
      ; l1 = (list 2)
      (define result (append l1 l2))
      ; result = (list 2 3 4)
      (cons h result)
    ]
  )
)

;(append (list 1 2) (list 3 4))



(define (foldr handle-step base-case l)
  (match l
    [(list)  base-case]
    [(list h l ...)
      ;(displayln (format "inside foldr: ~a" h))
      (define result (foldr handle-step base-case l))
      (handle-step h result)]))

(define (copy2 l)
  (foldr
    ; recursive step
    (lambda (h result)
      (displayln h)
      (cons h result)
    )
    ; base case
    (list)
    ; for each element in
    l
  )
)
;; (copy2 (list 1 2 3))
;
(define (map2 f l)
  (foldr
    ; recursive step (h result)
    (lambda (next-element recursive-result)
      (cons (f next-element) recursive-result)
    )
    ; base case
    (list)
    ; list
    l
  )
)

#|
(map2
  (lambda (x)
    (exact-floor (+ x 1))
  )
  (list 1.1 2.6 3.0))
|#

(define (append2 l1 l2)
  (foldr
    ; recursive step
    (lambda (h result) (cons h result))
    ; base-case
    l2
    ; list
    l1
  )
)
; (append2 (list 2 3) (list 4 5))
